/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef PAN_TREE_STORE_H
#define PAN_TREE_STORE_H

#include <deque>
#include <map>
#include <vector>
#include <gtk/gtk.h>

#if !GTK_CHECK_VERSION(2,6,0)
#define GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID (-2)
#endif

#define PAN_TREE_STORE_TYPE \
  (pan_tree_store_get_type ())
#define PAN_TREE_STORE(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST ((obj), PAN_TREE_STORE_TYPE, PanTreeStore))
#define PAN_TREE_STORE_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST ((klass),  PAN_TREE_STORE_TYPE, PanTreeStoreClass))
#define IS_PAN_TREE_STORE(obj)\
  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), PAN_TREE_STORE_TYPE))
#define IS_PAN_TREE_STORE__CLASS(klass)\
  (G_TYPE_CHECK_CLASS_TYPE ((klass),  PAN_TREE_STORE_TYPE))
#define PAN_TREE_STORE_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj),  PAN_TREE_STORE_TYPE, PanTreeStoreClass))

/**
 *  PanTreeStore is a replacement for GtkTreeStore aimed at scaling better with very
 *  large trees.  Much of this design came from looking at profiles of Pan's use of
 *  GtkTreeStore and desigining around the hotspots.  It has no other Pan dependencies
 *  and can be reused outside of Pan by copying the pan-tree.h and pan-tree.cc.
 * 
 *  1. Rows can be initialized when created.  This means no extra calls to
 *     tree_store_set_value(), no GObject type checking, no extra GtkTreePaths
 *     for row-changed, and most importantly, no row-changed signal at all.
 *
 *  2. Multiple rows can be inserted at once, which reduces some overhead in GObject
 *     type checking and GtkTreePath creation for the row-inserted signal handlers.
 *     It also allows us to allocate internal memory in chunks.
 *
 *  3. Children know what child number they are from their parent.  This adds an
 *     overhead of sizeof(int)*rows, but is far faster than the GtkTreeStore approach
 *     of having the children in linked lists where finding a child's index requires
 *     walking the linked list.
 *
 *  4. If you are setting the value of a cell, and the cell's GType matches the data
 *     type you passed in (such as calling celL_st_string() on a cell whose column
 *     type is G_TYPE_STRING), there are no intermediate GValue objects, thus no
 *     unnecessary g_strdup's or g_free's.
 * 
 *  5. Columns of GValues are allocated in arrays, rather than linked lists, saving us
 *     the overhead of ROW x COL GList nodes.
 *
 * TODO 

 *  -  There's currently some ugliness that needs to be cleaned up;
 *     in particular, way too much of PanTreeStore is currently public
 *     so that anonymous functions in pan-tree.cc can see them.
 *     For now, just pretend they're private.
 *
 *  -  Add a CellSetter API for setting contents even after they're added.
 *
 *  -  Sorting must be done manually after cells are changed.
 *
 *  -  Sorting must be done manually after cells are changed.
 */
struct PanTreeStore
{
  public:
    GObject parent;

  public:

    /**
     * The tree store passes as an argument to RowSetter::set_row()
     * for the RowSetter implemented by the user and passed as
     * an argument to pan_tree_store_{append,insert}_n().
     *
     * By setting the row before emitting a row-added signal,
     * no row-changed signal is needed.  The speedups resulting
     * from this are very substantial for large trees.
     *
     * Most of these are convenience functions around using a GValue.
     * But if you have a boxed type or something not covered in these
     * primitives, just use set_sell_gvalue().
     */
    struct CellSetter {
      virtual ~CellSetter () {}
      virtual bool set_cell_gvalue (int column, const GValue& value) = 0;
      virtual bool set_cell_int (int column, int value) = 0;
      virtual bool set_cell_uint (int column, unsigned int value) = 0;
      virtual bool give_cell_string (int column, char * value) = 0;
      virtual bool set_cell_string (int column, const char * value) = 0;
      virtual bool set_cell_pointer (int column, void* value) = 0;
      virtual bool set_cell_long (int column, long value) = 0;
      virtual bool set_cell_ulong (int column, unsigned long value) = 0;
      virtual bool set_cell_char (int column, char value) = 0;
      virtual bool set_cell_bool (int column, bool value) = 0;
    };

    /**
     * Base class for the class that the caller must implement to
     * call set_cell_*() for the various columns in a row.
     * the `n' in set_row refers to the row number being set,
     * in the range of [0...n_rows] where n_rows was passed into
     * pan_tree_store_{insert|append)_n().
     */
    struct RowSetter {
      virtual ~RowSetter() {}
      virtual void set_row (int n, CellSetter&) = 0;
    };

  public:

    /** This should really be private. */
    struct SortInfo {
      GtkTreeIterCompareFunc sort_func;
      gpointer user_data;
      GtkDestroyNotify destroy_notify;

      SortInfo(): sort_func(0), user_data(0), destroy_notify(0) {}
      ~SortInfo() { clear(); }
      void clear () {
        if (destroy_notify) destroy_notify (user_data);
        sort_func = 0;
        user_data = 0;
        destroy_notify = 0;
      }
      void assign (GtkTreeIterCompareFunc sort_func,
                   gpointer user_data,
                   GtkDestroyNotify destroy_notify) {
        clear ();
        this->sort_func = sort_func;
        this->user_data = user_data;
        this->destroy_notify = destroy_notify;
      }
    };
    /** sort info for any column */
    std::map<int,SortInfo> * sort_info;

    /** one of GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID,
               GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID,
               or [0...n_columns) */
    int sort_column_id; 

    /** the model's sort order -- ascending or descending */
    GtkSortType order;

  public:

    /**
     * Represents a row in a PanTreeStore.
     */
    struct Node
    {
      /** Pointer to the parent node. */
      Node * parent;
      /** Number of children this row has. */
      int n_children;
      /** Pointers to the child nodes. */
      std::vector<Node*> children;
      /** This node's index in it's parent's `children' vector.
          Used for fast building of GtkTreePath. */
      int child_index;
      /** Pointer to an array of GValues that make up
          this row's contents. */
      GValue * columns;

      Node(): parent(0), n_children(0), child_index(-1), columns(0) {}

      /** Note that the PanTreeStore needs to have unset the columns'
          values and free the column array before the destructor is called.
          This is because on its own the node doesn't know how many
          columns it has. */
      ~Node() {
        parent = (Node*) 0xDEADBEEF;
        children.clear ();
        n_children = 0xDEADBEEF;
        child_index = 0xDEADBEEF;
        columns = (GValue*) 0xDEADBEEF;
      }
    };

  public:

    /** Used just as in GtkTreeStore, to verify that GtkIters
        passed as arguments to us actually do belong to this tree. */
    int stamp;

    /** Defines the GType of each column. */    
    std::vector<GType> * column_types;

    /** A blank column. */
    GValue * columns_template;

    /** Number of columns in this tree. */
    int n_columns;

    /** The root node,  This is for implementation only;
        it's invisible to the outside world and you can't
        get a GtkTreeIter or GtkTreePath to it. */
    Node * root;

    typedef std::deque<Node> nodes_t;

    /** Acts as a memchunk of Nodes. */
    std::deque<Node> * nodes;

    /** Deleted nodes go here for future reuse */
    std::deque<Node*> * reuse_me;
};

struct PanTreeStoreClass
{
  GObjectClass parent_class;
};


/**********
***********
***********
**********/


PanTreeStore*  pan_tree_store_new   (int n_cols, ...);
GType     pan_tree_store_get_type   (void);
void      pan_tree_store_set        (PanTreeStore*, GtkTreeIter*, ...);
void      pan_tree_store_set_valist (PanTreeStore*, GtkTreeIter*, va_list);
gboolean  pan_tree_store_remove     (PanTreeStore*, GtkTreeIter*);
gpointer  pan_tree_store_peek_value (PanTreeStore*, GtkTreeIter*, int col);


/**
***  The fastest ways to add rows.
***  Append is marginally faster than insert.
**/

void      pan_tree_store_insert_n   (PanTreeStore             * tree,
                                     GtkTreeIter              * iter,
                                     GtkTreeIter              * parent,
                                     PanTreeStore::RowSetter  & row_setter,
                                     int                        n_rows,
                                     int                        pos);

void      pan_tree_store_append_n   (PanTreeStore             * tree,
                                     GtkTreeIter              * iter,
                                     GtkTreeIter              * parent,
                                     PanTreeStore::RowSetter  & row_setter,
                                     int                        n_rows);

/**
*** drop-in compatability with GtkTreeStore.
**/

void      pan_tree_store_append         (PanTreeStore * tree,
                                         GtkTreeIter  * iter,
                                         GtkTreeIter  * parent);

void      pan_tree_store_insert         (PanTreeStore * tree,
                                         GtkTreeIter  * iter,
                                         GtkTreeIter  * parent,
                                         gint           position);

void      pan_tree_store_insert_before  (PanTreeStore * tree,
                                         GtkTreeIter  * iter,
                                         GtkTreeIter  * parent,
                                         GtkTreeIter  * sibling);

void      pan_tree_store_insert_after   (PanTreeStore * tree,
                                         GtkTreeIter  * iter,
                                         GtkTreeIter  * parent,
                                         GtkTreeIter  * sibling);

void      pan_tree_store_prepend        (PanTreeStore * tree,
                                         GtkTreeIter  * iter,
                                         GtkTreeIter  * parent);
#endif
